from typing import Optional, Dict, List
import time
from datetime import datetime, timezone
import kubernetes
from kubernetes.client.rest import ApiException


class SAResource:
    def __init__(
        self,
        name: str,
        namespace: str,
        metadata: Optional[Dict] = None,
        extra_kwargs: Dict = {}, 
    ):
        self._resource_cache = None
        self._resource_token_secret_cache = None
        self.name = name
        self.namespace = namespace
        self.metadata = metadata if isinstance(metadata, dict) else {}
        self.automount_service_account_token = extra_kwargs.get(
            "automount_service_account_token", False)

    def get_text(self):
        """Return the text of the ServiceAccount that will be set to the
        kubernetes api."""
        metadata = self.metadata
        metadata.update({"name": self.name})
        return kubernetes.client.V1ServiceAccount(
            kind="ServiceAccount",
            metadata=kubernetes.client.V1ObjectMeta(**metadata),
            automount_service_account_token=self.automount_service_account_token,
        )

    def get_resource(self, api_client: kubernetes.client.ApiClient, cache=True):
        """Get the ServiceAccount object from the kubernetes cluster based on 
        the self.name. If cache is set to True, then cache the result and fetch from
        the cache on subsequent lookups.
        """
        if cache and self._resource_cache:
            return self._resource_cache
        api_instance = kubernetes.client.CoreV1Api(api_client)

        try:
            response = api_instance.read_namespaced_service_account(
                name=self.name, namespace=self.namespace)
        except ApiException as exc:
            response = None
            if exc.status != 404:
                raise
        self._resource_cache = response
        return response

    def resource_exists(
        self, api_client: kubernetes.client.ApiClient, cache: Optional[bool] = True
    ) -> bool:
        """Return if the ServiceAccount exists in the cluster"""
        return bool(self.get_resource(api_client, cache))

    def create(self, api_client: kubernetes.client.ApiClient):
        """Create the ServiceAccount in the kubernetes cluster"""
        if not self.resource_exists(api_client):
            api_instance = kubernetes.client.CoreV1Api(api_client)
            return api_instance.create_namespaced_service_account(
                namespace=self.namespace, body=self.get_text())
        else:
            return self.get_resource(api_client)

    def get_token_secret_resource_name(self, api_client: kubernetes.client.ApiClient, timeout: int = 10):
        start = time.time()
        cert = None
        while time.time() - start < timeout:
            sa = self.get_resource(api_client, cache=False)
            try:
                token = [s for s in sa.secrets if 'token' in s.name][0].name
            except (IndexError, AttributeError, TypeError):
                token = None
            if token:
                break
            time.sleep(1)
        return token

    def get_token_secret_resource(self, api_client: kubernetes.client.ApiClient, cache=True):
        if cache and self._resource_token_secret_cache:
            return self._resource_token_secret_cache
        api_instance = kubernetes.client.CoreV1Api(api_client)
        token_resource_name = self.get_token_secret_resource_name(api_client)
        try:
            response = api_instance.read_namespaced_secret(
                name=token_resource_name, namespace=self.namespace)
        except ApiException as exc:
            response = None
            if exc.status != 404:
                raise
        self._resource_token_secret_cache = response
        return response

    def get_token(self, api_client: kubernetes.client.ApiClient, timeout: int = 10):
        start = time.time()
        token = None
        while time.time() - start < timeout:
            secret = self.get_token_secret_resource(api_client, cache=False)
            try:
                token = secret.data['token']
            except (IndexError, AttributeError):
                token = None
            if token:
                break
            time.sleep(1)
        return token



import os
import sys
import argparse
import kubernetes
from kubernetes import client, config
from .user import CSRK8sUser, TokenK8sUser

def main(args=None):

    parser = argparse.ArgumentParser(
        prog="k8s_user", description="Create K8S User", epilog="")

    subparsers = parser.add_subparsers(help='', dest="user_type")

    parser.add_argument(
        "-k", "--out-kubeconfig",
        dest="out_kubeconfig",
        help=(
            "Output kubeconfig file path. A new kubeconfig "
            "file will be generated at this location with "
            "the credentials needed to login as the specified user."
        )
    )
    parser.add_argument(
        "--out-kubeconfig-context-name",
        dest="out_context",
        help=(
            "When used with the --out-kubeconfig option, this is the "
            "name of the kubeconfig context that will be associated with the user."
        ),
        default="default",
    )
    parser.add_argument(
        "--out-kubeconfig-cluster-name",
        dest="out_cluster",
        help=(
            "When used with the --out-kubeconfig option, this is the "
            "name of the kubeconfig cluster that will be associated with the user"
        ),
        default="default",
    )
    parser.add_argument(
        "--kubeconfig",
        dest="in_kubeconfig",
        help=(
            "This is the name of the kubeconfig that will be used to create "
            "the new user. This kubeconfig requires permissions to create, view, "
            "and approve k8s CSR resources. If not specified, the default kubeconfig "
            "will be used."
        ) ,
        default=None,
    )
    parser.add_argument(
        "-d",
        "--out-dir",
        dest="out_directory",
        help=(
            "If passed, the newly created public key, CSR, and cert files "
            "will be saved to this location in PEM format. This is in case "
            "you want to have these creds available in PEM format in "
            "addition to having the creds embedded in the kubeconfig file."
        ),
        default=None,
    )
    parser_csr = subparsers.add_parser('csr', help='CSR User Generator')
    parser_csr.add_argument(
        "name",
        nargs="?",
        help=(
            "The name of the new user to create, as well as the name "
            "of the k8s resource that will be attached to this user."
            )
        )

    parser_csr.add_argument(
        "--in-key",
        dest="in_key",
        help=(
            "Optionally pass in a filesystem path to an existing RSA key in PEM "
            "format instead of generating a new key."
        ),
        default=None,
    )
    parser_csr.add_argument(
        "--in-key-password",
        dest="in_key_password",
        help=(
            "If --in-key is passed, and if that key has a password set, "
            "this is the password to decrypt that key."
        ),
        default=None,
    )
    parser_csr.add_argument(
        "--in-csr",
        dest="in_csr",
        help=("Optonally pass in a filesystem path to an existing CSR file in PEM "
              "format. This requires the --in-key argument and requires that "
              "the CSR belongs to the KEY (the modulus match)."),
        default=None,
    )

    parser_token = subparsers.add_parser('sa', help='SA Token User Generator')
    parser_token.add_argument(
        "name",
        nargs="?",
        help=(
            "The name of the new user to create, as well as the name "
            "of the Service Account resource that will be attached to this user."
            )
        )
    parser_token.add_argument(
        "namespace",
        nargs="?",
        default='default',
        help=(
            "The namespace of the service account associated with the user."
            )
        )

    args = parser.parse_args()
    print(args)

    #sys.exit(1)

    if not args.user_type:
        print("user_type argument must be specified")
        sys.exit(1)

    if not args.name:
        print("Name argument must be specified")
        sys.exit(1)

    out_directory = args.out_directory

    if args.out_kubeconfig:
        out_kubeconfig = args.out_kubeconfig
    else:
        out_kubeconfig = f"{args.name}-kubeconfig.yaml"
        if out_directory:
            out_kubeconfig = os.path.join(out_directory, out_kubeconfig)

    if os.path.isfile(out_kubeconfig):
        print("kubeconfig file exists already at this location")
        sys.exit(1)

    api_client = config.new_client_from_config(config_file=args.in_kubeconfig)

    try:
        inputs_common = dict(
            cluster_name=args.out_cluster,
            context_name=args.out_context,
            out_kubeconfig=out_kubeconfig,
        )
        if args.user_type == "csr":
            user = CSRK8sUser(
                name=args.name,
            )

            inputs = {**dict(
                creds_dir=out_directory,
                in_key=args.in_key,
                in_key_password=args.in_key_password,
                in_csr=args.in_csr,
            ), **inputs_common}
        elif args.user_type == "sa":
            user = TokenK8sUser(
                name=args.name,
            )
            inputs = {**dict(
                namespace=args.namespace,
             ), **inputs_common}
        else:
            raise Exception("Must include a user_type as argument")
        user.create(api_client, inputs)
    except Exception as e:
        raise
        print(f"{e}")
        sys.exit(1)


if __name__ == "__main__":
    main()



import os
from typing import Dict
import base64
import collections
from abc import ABC, abstractmethod
import yaml
from .k8s.csr_resource import CSRResource
from .pki import Cert, CSRandKey, KeyBundle
from .kubeconfig import CSRKubeConfig, TokenKubeConfig
from .workflows.csr_workflow import UserCSRWorkflow
from .workflows.sa_workflow import UserTokenWorkflow


class K8sUser(ABC):

    def __init__(self, name):
        self.name = name

    def get_user_create_workflow_klass(self):
        return None 

    @abstractmethod
    def get_kubeconfig_klass(self):
        return None 

    def additional_inputs(self, inputs: Dict) -> Dict:
        return inputs

    def create(self, api_client, inputs: Dict) -> None:
        user_create_workflow_klass = self.get_user_create_workflow_klass()
        if not user_create_workflow_klass:
            raise NotImplementedError(
                "create method requires a get_user_create_workflow_klass")
        self.api_client = api_client

        user_create_workflow_klass(
            inputs={**dict(
                api_client=api_client,
                kubeconfig_klass=self.get_kubeconfig_klass(),
                user=self,
            ), **self.additional_inputs(inputs)},
        ).start()


class CSRK8sUser(K8sUser):

    def get_kubeconfig_klass(self):
        return CSRKubeConfig 

    def get_user_create_workflow_klass(self):
        return UserCSRWorkflow 


class TokenK8sUser(K8sUser):

    def get_kubeconfig_klass(self):
        return TokenKubeConfig 

    def get_user_create_workflow_klass(self):
        return UserTokenWorkflow 



import sys
import abc
from typing import Dict
import collections

StepReturn = collections.namedtuple(
    "StepReturn", "next_step message"
)

class BaseStep(abc.ABC):

    name = "base"

    def __init__(self, inputs: Dict):
        self.user = inputs.get('user')
        self.api_client = inputs.get('api_client')

    @abc.abstractmethod
    def run(self):
        return None

    def _run(self):
        print(f"Running: {self.name}", file=sys.stdout)
        step_return = self.run()
        print(f"  {step_return.message}", file=sys.stdout)
        return step_return


class EndStep(BaseStep):

    name="end"

    def run(self) -> StepReturn:
        return StepReturn(next_step=None, message="")


class WorkflowBase(abc.ABC):

    step_instances = {}

    def __init__(self, inputs: Dict):
        self.start_step = self.get_start_step().name
        for step_class in self.steps:
            self.step_instances[step_class.name] = step_class(inputs)

    @abc.abstractmethod
    def get_start_step(self):
        return None

    def start(self):
        step = StepReturn(self.start_step, '')
        while step.next_step:
            step = self.step_instances[step.next_step]._run()



import os
from unittest import mock
import json
import pytest
import kubernetes
from kubernetes.client.rest import ApiException
from k8s_user.k8s.sa_resource import SAResource


FIXTURE_DIR = os.path.join(
    os.path.dirname(os.path.realpath(__file__)),
    "..", 'fixtures',
    )


def test__saresource__init():
    name = "joe"
    namespace = "default"
    sar = SAResource(
        name=name,
        namespace=namespace,
    )
    with open(os.path.join(FIXTURE_DIR, "04_sa_resource_init.json")) as f:
        expected_sa_resource = json.loads(f.read())
        assert sar.get_text().to_dict() == expected_sa_resource


@mock.patch('k8s_user.k8s.sa_resource.kubernetes.client.CoreV1Api')
def test__saresource__get_resource(mock_CoreV1Api):

    name = "joe"
    namespace = "namespace"
    mock_response = {'dummy': "response"}
    mock_read_sa = mock.MagicMock()
    mock_read_sa.return_value = mock_response
    mock_CoreV1Api.return_value.read_namespaced_service_account = mock_read_sa

    mock_api_client = mock.Mock(spec=kubernetes.client.ApiClient) 
    sar = SAResource(
        name=name,
        namespace=namespace,
    )
    assert sar.get_resource(mock_api_client) == mock_response
    mock_read_sa.assert_called_once_with(name=name, namespace=namespace)
    # test the cache worked
    assert sar._resource_cache == mock_response


@mock.patch('k8s_user.k8s.sa_resource.kubernetes.client.CoreV1Api')
def test__saresource__get_resource__404(mock_CoreV1Api):

    name = "joe"
    namespace = "namespace"
    mock_response = {'dummy': "response"}
    mock_read_sa = mock.MagicMock()
    mock_read_sa.side_effect = ApiException(status = 404)
    mock_CoreV1Api.return_value.read_namespaced_service_account = mock_read_sa

    mock_api_client = mock.Mock(spec=kubernetes.client.ApiClient) 
    sar = SAResource(
        name=name,
        namespace=namespace,
    )
    assert sar.get_resource(mock_api_client) is None
    mock_read_sa.assert_called_once_with(name=name, namespace=namespace)


@pytest.mark.parametrize("resource,expected", 
[
    pytest.param({"dummy": "resource"}, True),
    pytest.param(None, False),
])
def test__saresource__resource_exists(resource, expected):
    name = "joe"
    namespace = "default"
    mock_api_client = mock.Mock(spec=kubernetes.client.ApiClient) 
    sar = SAResource(
        name=name,
        namespace=namespace)
    with mock.patch.object(sar, 'get_resource') as mock__get_resource:
        mock__get_resource.return_value = resource
        assert sar.resource_exists(mock_api_client) == expected


@mock.patch('k8s_user.k8s.sa_resource.kubernetes.client.CoreV1Api')
def test__saresource__create__new(mock_CoreV1Api):

    mock_response = {'dummy': "response"}
    mock_create_sa = mock.MagicMock()
    mock_create_sa.return_value = mock_response
    mock_CoreV1Api.return_value.create_namespaced_service_account = mock_create_sa

    name = "joe"
    namespace = "default"
    mock_api_client = mock.Mock(spec=kubernetes.client.ApiClient) 
    sar = SAResource(
        name=name,
        namespace=namespace)

    with mock.patch.object(sar, 'resource_exists') as mock__resource_exists:
        mock__resource_exists.return_value = False
        assert sar.create(mock_api_client) == mock_response
        mock_create_sa.assert_called_once()


@mock.patch('k8s_user.k8s.sa_resource.kubernetes.client.CoreV1Api')
def test__saresource__create__existing(mock_CoreV1Api):

    mock_response = {'dummy': "response"}

    name = "joe"
    namespace = "default"
    mock_api_client = mock.Mock(spec=kubernetes.client.ApiClient) 
    sar = SAResource(
        name=name,
        namespace=namespace)

    with mock.patch.object(sar, 'resource_exists') as mock__resource_exists:
        mock__resource_exists.return_value = True
        with mock.patch.object(sar, 'get_resource') as mock__get_resource:
            mock__get_resource.return_value = mock_response
            assert sar.create(mock_api_client) == mock_response


@mock.patch('k8s_user.k8s.sa_resource.kubernetes.client.CoreV1Api')
def test__saresource__get_token_secret_resource(mock_CoreV1Api):

    mock_response = {'dummy': "response"}
    mock_read_secret = mock.MagicMock()
    mock_read_secret.return_value = mock_response

    name = "joe"
    namespace = "default"
    mock_api_client = mock.Mock(spec=kubernetes.client.ApiClient) 
    sar = SAResource(
        name=name,
        namespace=namespace)
    with mock.patch.object(sar, 'get_token_secret_resource_name') as mock__get_token_secret_resource_name:
        mock__get_token_secret_resource_name.return_value = "joe-token-1234"
        mock_CoreV1Api.return_value.read_namespaced_secret = mock_read_secret
        assert sar.get_token_secret_resource(mock_api_client) == mock_response


@mock.patch('k8s_user.k8s.sa_resource.kubernetes.client.CoreV1Api')
def test__saresource__get_token(mock_CoreV1Api):


    class DummSecret:
        def __init__(self):
            self.data = {"token": "mytoken"}

    name = "joe"
    namespace = "default"
    mock_api_client = mock.Mock(spec=kubernetes.client.ApiClient) 
    sar = SAResource(
        name=name,
        namespace=namespace)
    with mock.patch.object(sar, 'get_token_secret_resource') as mock__get_token_secret_resource:
        mock__get_token_secret_resource.return_value = DummSecret()
        assert sar.get_token(mock_api_client) == "mytoken"


from typing import Optional, Dict, List
import base64
import yaml
import kubernetes


def _read_cert(path):
    with open(path, "rb") as f:
        return f.read()


def _base64(cert_data):
    return base64.b64encode(cert_data).decode("utf-8")


class GenericConfigGen:
    def __init__(self, left, right, **kwargs):
        self.left = left
        self.right = right

    def to_dict(self):
        left = self.left.to_dict() if hasattr(self.left, "to_dict") else self.left
        right = self.right.to_dict() if hasattr(self.right, "to_dict") else self.right
        return {**right, **left}

    def __or__(self, other):
        return GenericConfigGen(self, other)


class ClusterConfigGen:
    """Add server information to a kubeconfig"""
    def __init__(
            self,
            api_client: kubernetes.client.ApiClient,
            cluster_name: str,
            **kwargs,
            ):
        self.api_client = api_client
        self.cluster_name = cluster_name

    @property
    def cluster_ca_cert(self):
        return _base64(_read_cert(self.api_client.configuration.ssl_ca_cert))

    @property
    def host(self):
        return self.api_client.configuration.host

    def to_dict(self):
        return {
            "clusters": [
                {
                    "cluster": {
                        "certificate-authority-data": self.cluster_ca_cert,
                        "server": self.host,
                    },
                    "name": self.cluster_name,
                },
            ]
        }

    def __or__(self, other):
        return GenericConfigGen(self, other)


class CSRUserConfigGen:
    """Add User information to a kubeconfig"""
    def __init__(
            self, 
            api_client: kubernetes.client.ApiClient,
            keybundle,
            **kwargs):
        self.api_client = api_client
        self.keybundle = keybundle

    @property
    def cert(self):
        return _base64(_read_cert(self.api_client.configuration.cert_file))

    def to_dict(self):
        return {
            "users": [
                {
                    "name": self.keybundle.user_name,
                    "user": {
                        "client-certificate-data": self.keybundle.user_cert,
                        "client-key-data": self.keybundle.user_key,
                    },
                },
            ]
        }

    def __or__(self, other):
        return GenericConfigGen(self, other)


class TokenUserConfigGen:
    """Add a token to a kubeconfig"""
    def __init__(
            self,
            api_client: kubernetes.client.ApiClient,
            tokenbundle,
            **kwargs):
        self.api_client = api_client
        self.tokenbundle = tokenbundle

    def to_dict(self):
        return {
            "users": [
                {
                    "name": self.tokenbundle.user_name,
                    "user": {
                        "token": self.tokenbundle.user_token,
                    },
                },
            ]
        }

    def __or__(self, other):
        return GenericConfigGen(self, other)


class KubeConfigBase:

    def __init__(
            self,
            api_client: kubernetes.client.ApiClient,
            config_gen_klasses: Optional[Dict] = None,
            **kwargs):
        self.api_client = api_client
        self.kubeconfig_dict = {}
        self.config_kwargs = {}
        if config_gen_klasses:
            self._config_gen_klasses.update(config_gen_klasses)

    def generate(self):
        self.kubeconfig_dict = (
            self._config_gen_klasses["cluster"](**self.config_kwargs)
            | self._config_gen_klasses["user"](**self.config_kwargs)
            | self._config_gen_klasses["generic"](*self.static_config())
        ).to_dict()
        return self.kubeconfig_dict

    def save(self, path):
        if not self.kubeconfig_dict:
            self.generate()
        with open(path, "w") as f:
            f.write(yaml.dump(self.kubeconfig_dict))


class CSRKubeConfig(KubeConfigBase):
    """Generate a Kubeconfig with user csr"""

    _config_gen_klasses = {
        "cluster": ClusterConfigGen,
        "user": CSRUserConfigGen,
        "generic": GenericConfigGen,
    }

    def __init__(
        self,
        api_client: kubernetes.client.ApiClient,
        cluster_name: str,
        context_name: str,
        keybundle,
        config_gen_klasses: Optional[Dict] = None,
        **kwargs,
    ):
        super().__init__(
            api_client=api_client,
            config_gen_klasses=config_gen_klasses,
            **kwargs)

        updates = {
            "cluster_name": cluster_name,
            "context_name": context_name,
            "keybundle": keybundle,
            "api_client": api_client,
        }
        self.config_kwargs.update(updates)

    def static_config(self):
        return (
            {"apiVersion": "v1", "kind": "Config", "preferences": {}},
            {
                "current-context": self.config_kwargs['context_name'],
                "contexts": [
                    {
                        "context": {
                            "cluster": self.config_kwargs['cluster_name'],
                            "user": self.config_kwargs['keybundle'].user_name,
                        },
                        "name": self.config_kwargs['context_name'],
                    },
                ],
            }
        )


class TokenKubeConfig(KubeConfigBase):
    """Generate a Kubeconfig with user token"""

    _config_gen_klasses = {
        "cluster": ClusterConfigGen,
        "user": TokenUserConfigGen,
        "generic": GenericConfigGen,
    }

    def __init__(
        self,
        api_client: kubernetes.client.ApiClient,
        cluster_name: str,
        context_name: str,
        tokenbundle,
        config_gen_klasses: Optional[Dict] = None,
        **kwargs,
    ):
        super().__init__(
            api_client=api_client,
            config_gen_klasses=config_gen_klasses,
            **kwargs)

        updates = {
            "cluster_name": cluster_name,
            "context_name": context_name,
            "tokenbundle": tokenbundle,
            "api_client": api_client,
        }
        self.config_kwargs.update(updates)

    def static_config(self):
        return (
            {"apiVersion": "v1", "kind": "Config", "preferences": {}},
            {
                "current-context": self.config_kwargs['context_name'],
                "contexts": [
                    {
                        "context": {
                            "cluster": self.config_kwargs['cluster_name'],
                            "user": self.config_kwargs['tokenbundle'].user_name,
                        },
                        "name": self.config_kwargs['context_name'],
                    },
                ],
            }
        )



from typing import Optional, Dict, List
import time
from datetime import datetime, timezone
import kubernetes
from kubernetes.client.rest import ApiException


class CSRResource:
    """Class for managing the CertificateSigningRequest Kubernetes resource.
    
    :param name: the name of the CSR k8s resource
    :param csr_string: a base64 encoded CSR string
    :param metadata: an optional dict with fields matching k8s V1ObjectMeta object
    :param groups: RBAC groups to add to this CSR (defaults to ["system:authenticated"])
    :param usages: CSR usages (defaults to ["client auth"])
    """
    def __init__(
        self,
        name: str,
        csr_str: str,
        metadata: Optional[Dict] = None,
        groups: Optional[List] = None,
        usages: Optional[List] = None,
    ):
        self._resource_cache = None
        self.name = name
        self.csr_str = csr_str
        self.metadata = metadata if isinstance(metadata, dict) else {}
        self.groups = groups if groups else ["system:authenticated"]
        self.usages = usages if usages else ["client auth"]

    def get_text(self):
        """Return the text of the CertificateSigningRequest that will be set to the
        kubernetes api."""
        metadata = self.metadata
        metadata.update({"name": self.name})
        return kubernetes.client.V1beta1CertificateSigningRequest(
            kind="CertificateSigningRequest",
            metadata=kubernetes.client.V1ObjectMeta(**metadata),
            spec=kubernetes.client.models.v1beta1_certificate_signing_request_spec.V1beta1CertificateSigningRequestSpec(
                request=self.csr_str, groups=self.groups, usages=self.usages,
            ),
        )

    def get_resource(self, api_client: kubernetes.client.ApiClient, cache=True):
        """Get the CertificateSigningRequest object from the kubernetes cluster based on 
        the self.name. If cache is set to True, then cache the result and fetch from
        the cache on subsequent lookups.
        """
        if cache and self._resource_cache:
            return self._resource_cache
        api_instance = kubernetes.client.CertificatesV1beta1Api(api_client)
        try:
            response = api_instance.read_certificate_signing_request_status(self.name)
        except ApiException as exc:
            response = None
            if exc.status != 404:
                raise
        self._resource_cache = response
        return response

    def resource_exists(
        self, api_client: kubernetes.client.ApiClient, cache: Optional[bool] = True
    ) -> bool:
        """Return if the CertificateSigningRequest exists in the cluster"""
        return bool(self.get_resource(api_client, cache))

    def create(self, api_client: kubernetes.client.ApiClient):
        """Create the CertificateSigningRequest in the kubernetes cluster"""
        if not self.resource_exists(api_client):
            api_instance = kubernetes.client.CertificatesV1beta1Api(api_client)
            return api_instance.create_certificate_signing_request(self.get_text())
        else:
            return self.get_resource(api_client)

    def approve(
        self,
        api_client: kubernetes.client.ApiClient,
        message: str = "This certificate was approved by the Python Client.",
        reason: str = "ApprovedForUser",
    ):
        csr_status = self.get_resource(api_client, cache=False)
        # create an approval condition
        approval_condition = kubernetes.client.V1beta1CertificateSigningRequestCondition(
            last_update_time=datetime.now(timezone.utc).astimezone(),
            message=message,
            reason=reason,
            type="Approved",
        )

        # patch the existing `body` with the new conditions
        # you might want to append the new conditions to the existing ones
        csr_status.status.conditions = [approval_condition]
        api_instance = kubernetes.client.CertificatesV1beta1Api(api_client)
        response = api_instance.replace_certificate_signing_request_approval(
            self.name, csr_status
        )
        self._resource_cache = response
        return response

    def get_cert(self, api_client: kubernetes.client.ApiClient, timeout: int = 10):
        """Get the certificate from the CSR object"""
        start = time.time()
        cert = None
        while time.time() - start < timeout:
            csr_status = self.get_resource(api_client, cache=False)
            cert = csr_status.status.certificate
            if cert:
                break
            time.sleep(1)
        return cert



import os
import pytest
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.backends import default_backend
from cryptography.x509 import CertificateSigningRequest
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey
from k8s_user.pki import CSRandKey, CSR, Key, Cert

FIXTURE_DIR = os.path.join(
    os.path.dirname(os.path.realpath(__file__)),
    'fixtures',
    )

def test__key__init():
    k = Key()
    private_key = serialization.load_pem_private_key(
            k.pem,
            password=None,
            backend=default_backend()
        )
    assert k.key.private_numbers().q == private_key.private_numbers().q
    assert k.key.private_numbers().q == private_key.private_numbers().q


def test__key__load__from_file():
    k = Key(key_file=os.path.join(FIXTURE_DIR, "01_crypto_key.pem"))
    assert isinstance(k.key, RSAPrivateKey)
    assert (k.key.private_numbers().q ==
            144297355258033432961864249587566476417513395433306785930002598799109052058830086122084681049876654212847918251721493275668666135181801876080277008509389732804065133754794433634493195096633781919315478959132232587747522900893559497688267541882710926783531946921083827536107483831548399710096915589910119439697)
    assert (k.key.private_numbers().p ==
            155077597940994614685764714448196016481909530132221791492830636731403171936034167712684220338148449908312644390351813791630716591276752117103135699908938691335706645010018715863379137684598697481255169748674781389790109894698488651804364689336883691051597751748383115970680207826187946147682102388504221500771)


def test__key__generate():
    k = Key()
    assert not k.key_data
    assert not k.key_file
    assert isinstance(k.key, RSAPrivateKey)


def test__key__pem():
    k = Key()
    assert k.pem.decode("utf-8").startswith(
        "-----BEGIN RSA PRIVATE KEY-----")
    assert k.pem.decode("utf-8").endswith(
        "-----END RSA PRIVATE KEY-----\n")


def test__key__key_size():
    k = Key(key_size=2048)
    assert k.key.key_size == 2048


def test__key__save(tmp_path):
    k = Key()
    save_key = tmp_path / "key.pem"
    save_key.write_text("")
    k.save(path=save_key)

    with open(save_key) as f:
        k_from_file = serialization.load_pem_private_key(
                f.read().encode('utf-8'),
                password=None,
                backend=default_backend()
            )
        assert k.key.private_numbers().q == k_from_file.private_numbers().q
        assert k.key.private_numbers().q == k_from_file.private_numbers().q


def test__csr__init__key_match_csr_pubkey():
    k = Key()
    c = CSR(key=k, common_name="joe")
    assert (
        k.key.public_key().public_numbers().n ==
        c.csr.public_key().public_numbers().n
    )


def test__csr__load__from_file():
    k = Key(key_file=os.path.join(FIXTURE_DIR, "test04.key.pem"))
    cc = CSR(
        key=k,
        common_name="joe",
        csr_file=os.path.join(FIXTURE_DIR, "test04.csr.pem"))
    assert isinstance(cc.csr, CertificateSigningRequest)
    assert cc.csr.public_key().public_numbers().n == 47697849059556405225747884313444170230030763816433424631631373035711182936368092649968208614416434022706607776706324413502341646317728827547598067098252623604160016708295920244436964094624216727598081805820171341579513988264184396479085413667959684451333717475411285092373155692928873455930722519946733243885198520393782085591298880494985667255726797562061384487642807064782445358954937287712299058062046560400309940827255822634838060272875179078009131074978614249938179755360975280671537534506114111380505615383103391819800386789809167326284319333814600623689721087222405250502813469759449422767530797497646722237196915919525305165416484479205320285164625866942224505136347391509813145384639803900734490703721063552632775563762971875368305794470978647770107216064397414692148379270363411512853466501343962894822538634363330537679807995968768407745496270487574977770291585894697235327603924041831173934111279292064391522446476633459734055630905561810936892196777410004514290786614793639818503079345991842899845903465235429185309458399354358565380517620947176121344242378629202349277826925156970347574030013644291109396127155790643428605001831430439957635533484191537871485301724854443550268456964714674197789858655962524737938114211
    assert cc.csr.is_signature_valid


def test__csr__generate():
    k = Key()
    cc = CSR(
        key=k,
        common_name="joe",)
    assert not cc.csr_file
    assert not cc.csr_data
    assert isinstance(cc.csr, CertificateSigningRequest)
    assert cc.csr.is_signature_valid


@pytest.mark.parametrize("cn,additional_subject,expected", 
[
    pytest.param("joe", None, "CN=joe"),
    pytest.param("joe", {"O": "corp"}, "O=corp,CN=joe"),
    pytest.param("joe", {"OU": "IT"}, "OU=IT,CN=joe"),
    pytest.param("joe", {"C": "US"}, "C=US,CN=joe"),
    pytest.param("joe", {"ST": "IL"}, "ST=IL,CN=joe"),
    pytest.param("joe", {"L": "Chicago"}, "L=Chicago,CN=joe"),
    pytest.param("joe", {"SN": "Doe"}, "2.5.4.4=Doe,CN=joe"),
    pytest.param("joe", {"GN": "Joe"}, "2.5.4.42=Joe,CN=joe"),
    pytest.param("joe", {"T": "mytitle"}, "2.5.4.12=mytitle,CN=joe"),
    pytest.param("joe", {
        "O": "corp",
        "OU": "IT",
        "T": "mytitle",
        "C": "US",
        "ST": "IL",
        "L": "Chicago",
        "SN": "Doe",
        "GN": "Joe",
    }, "2.5.4.42=Joe,2.5.4.4=Doe,L=Chicago,ST=IL,C=US,2.5.4.12=mytitle,OU=IT,O=corp,CN=joe"),
])
def test__csr__subject(cn, additional_subject, expected):
    k = Key()
    c = CSR(key=k, common_name=cn, additional_subject=additional_subject)
    assert c.subject == expected


def test__csr__pem():
    k = Key()
    cc = CSR(key=k, common_name="joe")
    assert cc.pem.decode("utf-8").startswith(
        "-----BEGIN CERTIFICATE REQUEST-----")
    assert cc.pem.decode("utf-8").endswith(
        "-----END CERTIFICATE REQUEST-----\n")


def test__csr__base64():
    k = Key()
    cc = CSR(key=k, common_name="joe")
    assert cc.base64[:32] == "LS0tLS1CRUdJTiBDRVJUSUZJQ0FURSBS"


def test__csr__save(tmp_path):
    k = Key()
    cc = CSR(key=k, common_name="joe")
    save_csr = tmp_path / "csr.pem"
    save_csr.write_text("")
    cc.save(path=save_csr)

    with open(save_csr) as f:
        saved_csr = f.read()

    assert saved_csr.startswith(
        "-----BEGIN CERTIFICATE REQUEST-----")
    assert saved_csr.endswith(
        "-----END CERTIFICATE REQUEST-----\n")


def test__cert__init__crt_file():
    ct = Cert(crt_file=os.path.join(FIXTURE_DIR, "03_cert.pem"))
    assert ct.crt.serial_number == 61276984187087310175771381080539889888


def test__cert__init__crt_data():
    with open(os.path.join(FIXTURE_DIR, "03_cert.pem"), 'rb') as f:
        ct = Cert(crt_data=f.read())
    assert ct.crt.serial_number == 61276984187087310175771381080539889888


def test__cert__init__subject():
    ct = Cert(crt_file=os.path.join(FIXTURE_DIR, "03_cert.pem"))
    assert ct.subject == "CN=john2,O=jazstudios"


def test__cert__pem():
    ct = Cert(crt_file=os.path.join(FIXTURE_DIR, "03_cert.pem"))
    assert ct.pem.decode("utf-8").startswith(
        "-----BEGIN CERTIFICATE-----")
    assert ct.pem.decode("utf-8").endswith(
        "-----END CERTIFICATE-----\n")

def test__cert__save(tmp_path):
    ct = Cert(crt_file=os.path.join(FIXTURE_DIR, "03_cert.pem"))
    save_crt = tmp_path / "cert.pem"
    save_crt.write_text("")
    ct.save(path=save_crt)

    with open(save_crt) as f:
        saved_crt = f.read()
    assert saved_crt.startswith(
        "-----BEGIN CERTIFICATE-----")
    assert saved_crt.endswith(
        "-----END CERTIFICATE-----\n")


def test__csrandkey__init():
    ck = CSRandKey(
        common_name="joe",
        additional_subject={"O": "myorg"},
        key_size=2048,
    )
    assert ck.csr.subject == "O=myorg,CN=joe"

    assert (
        ck.key.key.public_key().public_numbers().n ==
        ck.csr.csr.public_key().public_numbers().n
    )



from typing import Optional, Dict, List, Any
from datetime import datetime, timezone
import base64
import collections
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.x509.oid import NameOID
from cryptography.hazmat.primitives import serialization


NAME_ATTRIBUTE_MAPING = {
    "O": NameOID.ORGANIZATION_NAME,
    "OU": NameOID.ORGANIZATIONAL_UNIT_NAME,
    "CN": NameOID.COMMON_NAME,
    "C": NameOID.COUNTRY_NAME,
    "ST": NameOID.STATE_OR_PROVINCE_NAME,
    "L": NameOID.LOCALITY_NAME,
    "STREET": NameOID.STREET_ADDRESS,
    "SN": NameOID.SURNAME,
    "GN": NameOID.GIVEN_NAME,
    "T": NameOID.TITLE,
    "DC": NameOID.DOMAIN_COMPONENT,
    "UID": NameOID.USER_ID,
}


KeyBundle = collections.namedtuple("KeyBundle", "user_name user_key user_csr user_cert")


class Key:
    """A wrapper object to manage a Key

    This can be used to generate a Key or it can be used to load
    and act upon an existing Key.
    """
    def __init__(
        self,
        key_size: int = 4092,
        key_data: Optional[bytes] = None,
        key_file: Optional[str] = None,
        key_file_password: Optional[str] = None,
    ):
        """
        :param key_size: the size of the RSA key generated. Defaults to 4092
        :param key_data: an optional bytestring of an existing PEM key. If provided,
            a new key will not be generated.
        :param key_file: an optional filesystem path to an existing PEM key. If provided,
            a new key will not be generated.
        :param key_password: an optional encryption password for the key_file if the key_file
            attribute is provided.
        """
        self.created = False
        self.key_size = key_size
        self.key_data = key_data
        self.key_file = key_file
        self.key_file_password = key_file_password

        if self.key_file or self.key_data:
            self.load()
        else:
            self.generate()
            self.created = True

    def load_data(self, key_data: bytes, password: Optional[str] = None):
        """Load a Key from a PEM-formatted byte string
        
        :param key_data: a byte string of Key data in PEM format
        :returns: a key object
        """
        private_key = serialization.load_pem_private_key(
            key_data, password=password, backend=default_backend()
        )
        return private_key

    def load_file(self, path: str, password: Optional[str] = None):
        """Load a Key from a PEM file
        
        :param path: a filesystem path to a PEM Key file
        :param password: an optional password to decrypt the key
        :returns: a key object
        """
        with open(path, "rb") as f:
            return self.load_data(f.read(), password=password)

    def load(self):
        """Loads a Key object based on key_data or key_file attribute"""
        if self.key_data:
            self.key = self.load_data(
                key_data=self.key_data, password=self.key_file_password)
        elif self.key_file:
            self.key = self.load_file(
                path=self.key_file, password=self.key_file_password)
        else:
            raise ValueError("Must supply key_data or key_file")

    def generate(self):
        """Generate a Key based on this object's attributes"""
        self.key = rsa.generate_private_key(
            public_exponent=65537, key_size=self.key_size, backend=default_backend()
        )

    @property
    def base64(self):
        """Return a base64-encoded representation of this Key."""
        return base64.b64encode(self.pem).decode("utf-8")

    @property
    def pem(self):
        """Return a PEM representation of this Key."""
        return self.key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.TraditionalOpenSSL,
            encryption_algorithm=serialization.NoEncryption(),
        )

    def save(self, path: str):
        """Save a PEM representation of this Key to the provided path"""
        with open(path, "wb") as f:
            f.write(self.pem)


class CSR:
    """A wrapper object to manage a CSR

    This can be used to generate a CSR or it can be used to load
    and act upon an existing CSR.
    """
    def __init__(
        self,
        key: Key,
        common_name: str,
        additional_subject: Optional[Dict] = None,
        dnsnames: Optional[Dict] = None,
        csr_data: Optional[bytes] = None,
        csr_file: Optional[str] = None,
        signing_hash_algo: Optional[Any] = None,
    ):
        """
        :param key: a pki.Key object to associate with this CSR.
        :param common_name: The Common Name (CN) for the certificate
        :param additional_subject: a dictionary where the keys are the x509 subject
            attributes from cryptography.x509.NameOID OR a short name abbreviation
            for those attributes. See the NAME_ATTRIBUTE_MAPING above.
        :param dnsnames: a list of Subject Alternative Name (SAN) DNS names.
        :param csr_data: an optional bytestring containing an existing PEM CSR. If
            this is specified, a key must be provided as well that matches this CSR.
            If this is given, a CSR will not be generated, but the provided on will be
            used.
        :param csr_file: an optional filesystem path to an existing PEM CSR. If this
            is specified, a key must be provided as well that matches this CSR.
            If this is given, a CSR will not be generated, but the provided one will be
            used.
        :param signing_hash_algo: a hash algorithm to sign the key with. Default is
            hashes.SHA256
        """
        self.created = False
        self.key = key
        self.csr_data = csr_data
        self.csr_file = csr_file
        self.signing_hash_algo  = (
            signing_hash_algo if signing_hash_algo else hashes.SHA256)

        if not additional_subject:
            additional_subject = {}
        self.attribue_list = [x509.NameAttribute(NameOID.COMMON_NAME, common_name)] + [
            x509.NameAttribute(NAME_ATTRIBUTE_MAPING.get(a, a), v)
            for a, v in additional_subject.items()
        ]

        if not dnsnames:
            dnsnames = {}
        self.dnsnames = dnsnames

        if self.csr_file or self.csr_data:
            self.load()
        else:
            self.generate()
            self.created = True

    def load_data(self, csr_data: bytes):
        """Load a CSR from a PEM-formatted byte string
        
        :param csr_data: a byte string of CSR data in PEM format
        :returns: a csr object
        """
        csr = x509.load_pem_x509_csr(
            csr_data, backend=default_backend()
        )
        return csr

    def load_file(self, path: str):
        """Load a CSR from a PEM file
        
        :param path: a filesystem path to a PEM CSR file
        :returns: a csr object
        """
        with open(path, "rb") as f:
            return self.load_data(f.read())

    def load(self):
        """Loads a CSR object based on csr_data or csr_file attribute"""
        if self.csr_data:
            self.csr = self.load_data(csr_data=self.csr_data)
        elif self.csr_file:
            self.csr = self.load_file(path=self.csr_file)
        else:
            raise ValueError("Must supply csr_data or csr_file")

    def generate(self):
        """Generate a CSR based on this object's attributes"""
        self.csr = (
            x509.CertificateSigningRequestBuilder()
            .subject_name(x509.Name(self.attribue_list))
            .add_extension(x509.SubjectAlternativeName(self.dnsnames), critical=False,)
            .sign(self.key.key, self.signing_hash_algo(), default_backend())
        )

    @property
    def subject(self):
        """Return a subject string based on this object's attributes"""
        return self.csr.subject.rfc4514_string()

    @property
    def pem(self):
        """Return a PEM representation of this CSR."""
        return self.csr.public_bytes(serialization.Encoding.PEM)

    @property
    def base64(self):
        """Return a base64-encoded representation of this CSR."""
        return base64.b64encode(self.pem).decode("utf-8")

    def save(self, path: str):
        """Save a PEM representation of this CSR to the provided path"""
        with open(path, "wb") as f:
            f.write(self.pem)


class Cert:
    """A wrapper object to manage a Cert

    This can be used to load and act upon an existing Cert.
    """
    def __init__(
        self, crt_data: Optional[bytes] = None, crt_file: Optional[str] = None
    ):
        """
        :param crt_data: an optional bytestring containing an existing PEM Cert
        :param crt_file: an optional filesystem path to an existing PEM Cert.
            used.
        """
        self.load(crt_data, crt_file)

    def load_data(self, crt_data: bytes):
        """Load a Cert from a PEM-formatted byte string
        
        :param crt_data: a byte string of Cert data in PEM format
        :returns: a cert object
        """
        return x509.load_pem_x509_certificate(crt_data, default_backend())

    def load_file(self, path: str):
        """Load a Cert from a PEM file
        
        :param path: a filesystem path to a PEM Cert file
        :returns: a cert object
        """
        with open(path, "rb") as f:
            return self.load_data(f.read())

    def load(self, crt_data=None, crt_file=None):
        """Loads a Cert object based on crt_data or crt_file attribute"""
        if crt_data:
            self.crt = self.load_data(crt_data)
        elif crt_file:
            self.crt = self.load_file(path=crt_file)
        else:
            self.crt = None

    @property
    def pem(self):
        """Return a PEM representation of this Cert."""
        return self.crt.public_bytes(serialization.Encoding.PEM)

    @property
    def base64(self):
        """Return a base64-encoded representation of this Cert."""
        return base64.b64encode(self.pem).decode("utf-8")

    @property
    def subject(self):
        """Return a subject string based on this object's attributes"""
        return self.crt.subject.rfc4514_string()

    def save(self, path: str):
        """Save a PEM representation of this Cert to the provided path"""
        with open(path, "wb") as f:
            f.write(self.pem)


class CSRandKey:
    """A wrapper object to hold both a CSR and Key

    This can be used to generate a Key and CSR, or it can be used to load
    and act upon an existing Key and CSR.
    """
    def __init__(
        self,
        common_name,
        additional_subject: Optional[Dict] = None,
        dnsnames: Optional[Dict] = None,
        key_size: int = 4092,
        key_file: Optional[str] = None,
        key_file_password: Optional[str] = None,
        csr_file: Optional[str] = None,
    ):
        """
        :param common_name: The Common Name (CN) for the certificate
        :param additional_subject: a dictionary where the keys are the x509 subject
            attributes from cryptography.x509.NameOID OR a short name abbreviation
            for those attributes. See the NAME_ATTRIBUTE_MAPING above.
        :param dnsnames: a list of Subject Alternative Name (SAN) DNS names.
        :param key_size: the size of the RSA key generated. Defaults to 4092
        :param key_file: an optional filesystem path to an existing PEM RSA key.
            Only use this when you wish to use an existing key. If you want
            to generate a new one, leave this empty.
        :param key_password: an optional encryption password for the key_file if the key_file
            attribute is provided.
        :param csr_file: an optional filesystem path to an existing PEM CSR. If this
            is specified, a key_file must be provided as well that matches this CSR.
            If this is given, a CSR will not be generated, but the provided one will be
            used.
        """
        self.key = Key(
            key_size=key_size,
            key_file=key_file,
            key_file_password=key_file_password,
        )
        self.csr = CSR(
            key=self.key,
            common_name=common_name,
            additional_subject=additional_subject,
            dnsnames=dnsnames,
            csr_file=csr_file,
        )



import collections
from . import StepReturn, BaseStep, EndStep, WorkflowBase
from ..k8s.sa_resource import SAResource


TokenBundle = collections.namedtuple(
    "TokenBundle", "user_name user_token")


class ResourceExistsStep(BaseStep):

    name="sa_resource_exists"

    def __init__(self, inputs):
        self.name = inputs.get("name")
        self.namespace = inputs.get("namespace")
        self.metadata = inputs.get("metadata")
        super().__init__(inputs)

    def run(self) -> StepReturn:

        self.user.sa_resource = SAResource(
            name=self.user.name,
            namespace=self.namespace,
            metadata=self.metadata,
        )
        exists = self.user.sa_resource.resource_exists(self.api_client)
        return StepReturn(
            next_step='sa_get_or_create_resource',
            message="resouce exists" if exists else "resource does not exist")


class GetorCreateSAStep(BaseStep):

    name="sa_get_or_create_resource"

    def run(self) -> StepReturn:
        self.user.sa_resource.create(self.api_client)
        return StepReturn(
            next_step='get_token',
            message="token retrieved")


class GetTokenStep(BaseStep):

    name="get_token"

    def run(self) -> StepReturn:
        token_str = self.user.sa_resource.get_token(self.api_client)
        self.user.token = token_str
        return StepReturn(
            next_step='make_kubeconfig',
            message="token generated")


class MakeKubeConfigStep(BaseStep):

    name="make_kubeconfig"

    def __init__(self, inputs):
        self.cluster_name = inputs.get("cluster_name")
        self.context_name = inputs.get("context_name")
        self.kubeconfig_klass = inputs.get("kubeconfig_klass")
        super().__init__(inputs)

    def run(self) -> StepReturn:

        tokenbundle = TokenBundle(
            user_name=self.user.name,
            user_token=self.user.token,
        )
        self.user.kubeconfig = self.kubeconfig_klass(
            self.api_client, self.cluster_name, self.context_name,
            tokenbundle,
        )
        self.user.kubeconfig_dict = self.user.kubeconfig.generate()

        return StepReturn(
            next_step='save_kubeconfig',
            message="kubeconfig generated")


class SaveKubeconfigStep(BaseStep):

    name="save_kubeconfig"

    def __init__(self, inputs):
        self.out_kubeconfig = inputs.get("out_kubeconfig")
        super().__init__(inputs)

    def run(self) -> StepReturn:
        self.user.kubeconfig.save(self.out_kubeconfig)
        return StepReturn(
            next_step='end',
            message=f"kubeconfig saved to {self.out_kubeconfig}")


class UserTokenWorkflow(WorkflowBase):

    steps = [
        ResourceExistsStep,
        GetorCreateSAStep,
        MakeKubeConfigStep,
        GetTokenStep,
        SaveKubeconfigStep,
        EndStep,
    ]

    def get_start_step(self):
        return ResourceExistsStep


import os
from typing import Dict
import collections
import base64
from ..pki import Cert, CSRandKey, KeyBundle
from ..k8s.csr_resource import CSRResource
from . import StepReturn, BaseStep, EndStep, WorkflowBase


class GetCSRandKeyStep(BaseStep):

    name="get_csr_and_key"

    def __init__(self, inputs):
        self.in_key = inputs.get("in_key")
        self.in_key_password = inputs.get("in_key_password")
        self.in_csr = inputs.get("in_csr")
        self.metadata = inputs.get("metadata")
        super().__init__(inputs)

    def run(self) -> StepReturn:
        self.user.candk = CSRandKey(
            common_name=self.user.name,
            key_file=self.in_key,
            key_file_password=self.in_key_password,
            csr_file=self.in_csr,
        )
        self.user.csr_resource = CSRResource(
            name=self.user.name,
            csr_str=self.user.candk.csr.base64,
            metadata=self.metadata
        )
        return StepReturn(
            next_step='save_key',
            message=(
                f"key {'created' if self.user.candk.key.created else 'loaded'}; "
                f"csr {'created' if self.user.candk.csr.created else 'loaded'}"))


class SaveKeyStep(BaseStep):

    name="save_key"

    def __init__(self, inputs):
        self.in_key = inputs.get("in_key")
        self.creds_dir = inputs.get("creds_dir")
        super().__init__(inputs)

    def run(self) -> StepReturn:
        saved = False
        if self.creds_dir and not self.in_key:
            key_path = os.path.join(self.creds_dir, f"{self.user.name}.key.pem")
            if os.path.exists(key_path):
                raise Exception(f"Key already exists at {key_path}")
            self.user.candk.key.save(key_path)
            saved = True
        return StepReturn(
            next_step='save_csr',
            message=f"key saved to {key_path}" if saved else "skipped save")


class SaveCSRStep(BaseStep):

    name="save_csr"

    def __init__(self, inputs):
        self.in_csr = inputs.get("in_csr")
        self.creds_dir = inputs.get("creds_dir")
        super().__init__(inputs)

    def run(self) -> StepReturn:
        saved = False
        if self.creds_dir and not self.in_csr:
            csr_path = os.path.join(self.creds_dir, f"{self.user.name}.csr.pem")
            self.user.candk.csr.save(csr_path)
            saved = True
        return StepReturn(
            next_step='csr_resource_exists',
            message=f"csr saved to {csr_path}" if saved else "skipped save")


class ResourceExistsStep(BaseStep):

    name="csr_resource_exists"

    def run(self) -> StepReturn:
        exists = self.user.csr_resource.resource_exists(self.api_client)
        if exists:
            return StepReturn(
                next_step='csr_approve_resource',
                message="csr resource exists")
        return StepReturn(
            next_step='csr_create_resource',
            message="csr resource does not exist yet.")


class CreateResourceStep(BaseStep):

    name="csr_create_resource"

    def run(self) -> StepReturn:
        self.user.csr_resource.create(self.api_client)
        return StepReturn(
            next_step='csr_approve_resource',
            message="csr resource created")


class ApproveResourceStep(BaseStep):

    name="csr_approve_resource"

    def run(self) -> StepReturn:
        self.user.csr_resource.approve(self.api_client)
        return StepReturn(
            next_step='get_cert',
            message="csr resource approved")


class GetCertStep(BaseStep):

    name="get_cert"

    def run(self) -> StepReturn:
        cert_str = self.user.csr_resource.get_cert(self.api_client)
        self.user.crt = Cert(crt_data=base64.b64decode(cert_str))
        return StepReturn(
            next_step='save_cert',
            message="crt retrieved from k8s")


class SaveCertStep(BaseStep):

    name="save_cert"

    def __init__(self, inputs):
        self.creds_dir = inputs.get("creds_dir")
        super().__init__(inputs)

    def run(self) -> StepReturn:
        saved = False
        if self.creds_dir:
            crt_path = os.path.join(self.creds_dir, f"{self.user.name}.crt.pem")
            self.user.crt.save(crt_path)
            saved = True
        return StepReturn(
            next_step='make_kubeconfig',
            message=f"crt saved to {crt_path}" if saved else "skipped save")


class MakeKubeConfigStep(BaseStep):

    name="make_kubeconfig"

    def __init__(self, inputs):
        self.cluster_name = inputs.get("cluster_name")
        self.context_name = inputs.get("context_name")
        self.kubeconfig_klass = inputs.get("kubeconfig_klass")
        super().__init__(inputs)

    def run(self) -> StepReturn:

        keybundle = KeyBundle(
            user_name=self.user.name,
            user_key=self.user.candk.key.base64,
            user_csr=self.user.candk.key.base64,
            user_cert=self.user.crt.base64,
        )
        self.user.kubeconfig = self.kubeconfig_klass(
            self.api_client, self.cluster_name, self.context_name, keybundle,
        )
        self.user.kubeconfig_dict = self.user.kubeconfig.generate()

        return StepReturn(
            next_step='save_kubeconfig',
            message="kubeconfig generated")


class SaveKubeconfigStep(BaseStep):

    name="save_kubeconfig"

    def __init__(self, inputs):
        self.out_kubeconfig = inputs.get("out_kubeconfig")
        super().__init__(inputs)

    def run(self) -> StepReturn:
        self.user.kubeconfig.save(self.out_kubeconfig)
        return StepReturn(
            next_step='end',
            message=f"kubeconfig saved to {self.out_kubeconfig}")


class UserCSRWorkflow(WorkflowBase):
    steps = [
        GetCSRandKeyStep,
        SaveKeyStep,
        SaveCSRStep,
        ResourceExistsStep,
        CreateResourceStep,
        ApproveResourceStep,
        GetCertStep,
        SaveCertStep,
        MakeKubeConfigStep,
        SaveKubeconfigStep,
        EndStep,
    ]

    def get_start_step(self):
        return GetCSRandKeyStep


import os
from unittest import mock
import pytest
import kubernetes
from k8s_user.pki import KeyBundle
from k8s_user.kubeconfig import (
    GenericConfigGen, ClusterConfigGen,
    CSRUserConfigGen, CSRKubeConfig, TokenKubeConfig)
from k8s_user.workflows.sa_workflow import TokenBundle


FIXTURE_DIR = os.path.join(
    os.path.dirname(os.path.realpath(__file__)),
    'fixtures',
    )


@pytest.mark.parametrize("in1,in2,expected", 
[
    pytest.param({}, {}, {}),
    pytest.param(
        {"name": "1"}, {"name2": "2"},
        {"name": "1", "name2": "2"}),
    pytest.param(
        {"name": {}}, {"name2": {}},
        {"name": {}, "name2": {}}),
    pytest.param(
        {"name": {"value": "1"}}, {"name2": {"a": "b"}},
        {"name": {"value": "1"}, "name2": {"a": "b"}}),
])
def test__GenericConfigGen(in1, in2, expected):
    assert GenericConfigGen(in1, in2).to_dict() == expected


@pytest.mark.parametrize("in1,in2,expected", 
[
    pytest.param({}, {}, {}),
    pytest.param(
        {"name": "1"}, {"name2": "2"},
        {"name": "1", "name2": "2"}),
    pytest.param(
        {"name": {}}, {"name2": {}},
        {"name": {}, "name2": {}}),
    pytest.param(
        {"name": {"value": "1"}}, {"name2": {"a": "b"}},
        {"name": {"value": "1"}, "name2": {"a": "b"}}),
])
def test__GenericConfigGen_2x(in1, in2, expected):
    assert (
        GenericConfigGen(in1, {}) |
        GenericConfigGen(in2, {})
    ).to_dict() == expected


def test__ClusterConfigGen(tmp_path):

    save_ca_cert = tmp_path / "ca-cert.pem"
    save_ca_cert.write_text("hi")

    class DummyConfiguration:
        host = "localhost"
        ssl_ca_cert = save_ca_cert

    mock_api_client = mock.Mock(spec=kubernetes.client.ApiClient)
    mock_api_client.configuration = DummyConfiguration()
    ccg = ClusterConfigGen(mock_api_client, "my_cluster")
    assert ccg.to_dict() == {
        'clusters': [{'cluster': {'certificate-authority-data': 'aGk=',
                                   'server': 'localhost'},
                      'name': 'my_cluster'}]}


def test__CSRUserConfigGen(tmp_path):

    save_cert = tmp_path / "cert.pem"
    save_cert.write_text("hi")

    dummy_keybundle = KeyBundle(
            user_name="myname",
            user_key="mykey",
            user_csr="mycsr",
            user_cert="mycrt",
        )

    class DummyConfiguration:
        host = "localhost"
        ssl_ca_cert = save_cert

    mock_api_client = mock.Mock(spec=kubernetes.client.ApiClient)
    mock_api_client.configuration = DummyConfiguration()
    ccg = CSRUserConfigGen(mock_api_client, dummy_keybundle)
    assert ccg.to_dict() == {
        'users': [{'name': 'myname',
                   'user': {'client-certificate-data': 'mycrt',
                            'client-key-data': 'mykey'}}]}


def test__multi__ConfigGen(tmp_path):

    save_cert = tmp_path / "cert.pem"
    save_cert.write_text("hi")

    save_ca_cert = tmp_path / "ca-cert.pem"
    save_ca_cert.write_text("hi")

    dummy_keybundle = KeyBundle(
            user_name="myname",
            user_key="mykey",
            user_csr="mycsr",
            user_cert="mycrt",
        )

    class DummyConfiguration:
        host = "localhost"
        ssl_ca_cert = save_cert

    mock_api_client = mock.Mock(spec=kubernetes.client.ApiClient)
    mock_api_client.configuration = DummyConfiguration()

    assert (ClusterConfigGen(mock_api_client, "my_cluster") |
            CSRUserConfigGen(mock_api_client, dummy_keybundle) |
            GenericConfigGen(
                {"apiVersion": "v1",
                "kind": "Config", "preferences": {}}, {})
            ).to_dict() == {
                'apiVersion': 'v1',
                'clusters': [{'cluster': {'certificate-authority-data': 'aGk=',
                                        'server': 'localhost'},
                            'name': 'my_cluster'}],
                'kind': 'Config',
                'preferences': {},
                'users': [{'name': 'myname',
                        'user': {'client-certificate-data': 'mycrt',
                                    'client-key-data': 'mykey'}}],
            }


def test__CSRKubeConfig(tmp_path):

    save_cert = tmp_path / "cert.pem"
    save_cert.write_text("hi")

    dummy_keybundle = KeyBundle(
            user_name="myname",
            user_key="mykey",
            user_csr="mycsr",
            user_cert="mycrt",
        )

    class DummyConfiguration:
        host = "localhost"
        ssl_ca_cert = save_cert

    mock_api_client = mock.Mock(spec=kubernetes.client.ApiClient)
    mock_api_client.configuration = DummyConfiguration()

    kc = CSRKubeConfig(
        mock_api_client,
        "mycontext",
        "mycluster",
        dummy_keybundle,
    )

    assert kc.generate() == {
        'apiVersion': 'v1',
        'clusters': [{'cluster': {'certificate-authority-data': 'aGk=',
                                  'server': 'localhost'},
                      'name': 'mycontext'}],
        'contexts': [{'context': {'cluster': 'mycontext',
                                  'user': 'myname'},
                      'name': 'mycluster'}],
        'current-context': 'mycluster',
        'kind': 'Config',
        'preferences': {},
        'users': [{'name': 'myname',
                   'user': {'client-certificate-data': 'mycrt',
                            'client-key-data': 'mykey'}}],
        }


def test__TokenKubeConfig(tmp_path):

    save_cert = tmp_path / "cert.pem"
    save_cert.write_text("hi")

    dummy_tokenbundle = TokenBundle(
            user_name="myname",
            user_token="mytoken",
        )

    class DummyConfiguration:
        host = "localhost"
        ssl_ca_cert = save_cert

    mock_api_client = mock.Mock(spec=kubernetes.client.ApiClient)
    mock_api_client.configuration = DummyConfiguration()

    kc = TokenKubeConfig(
        mock_api_client,
        "mycontext",
        "mycluster",
        dummy_tokenbundle,
    )

    assert kc.generate() == {
        'apiVersion': 'v1',
        'clusters': [{'cluster': {'certificate-authority-data': 'aGk=',
                                  'server': 'localhost'},
                      'name': 'mycontext'}],
        'contexts': [{'context': {'cluster': 'mycontext',
                                  'user': 'myname'},
                      'name': 'mycluster'}],
        'current-context': 'mycluster',
        'kind': 'Config',
        'preferences': {},
        'users': [{'name': 'myname',
                   'user': {'token': 'mytoken'}}],
        }



import os
from unittest import mock
import json
import pytest
import kubernetes
from kubernetes.client.rest import ApiException
from k8s_user.pki import CSRandKey, CSR, Key
from k8s_user.k8s.csr_resource import CSRResource


FIXTURE_DIR = os.path.join(
    os.path.dirname(os.path.realpath(__file__)),
    "..", 'fixtures',
    )


def test__csrresource__init():
    name = "joe"
    cak = CSRandKey(common_name=name)
    csrr = CSRResource(
        name=name,
        csr_str=cak.csr.base64)
    with open(os.path.join(FIXTURE_DIR, "02_csr_resource_init.json")) as f:
        expected_csr_resource = json.loads(f.read())
        expected_csr_resource['spec']['request'] = cak.csr.base64
        assert csrr.get_text().to_dict() == expected_csr_resource



@mock.patch('k8s_user.k8s.csr_resource.kubernetes.client.CertificatesV1beta1Api')
def test__csrresource__get_resource(mock_CertificatesV1beta1Api):

    name = "joe"
    mock_response = {'dummy': "response"}
    mock_read_csr = mock.MagicMock()
    mock_read_csr.return_value = mock_response
    mock_CertificatesV1beta1Api.return_value.read_certificate_signing_request_status = mock_read_csr

    mock_api_client = mock.Mock(spec=kubernetes.client.ApiClient) 
    cak = CSRandKey(common_name=name)
    csrr = CSRResource(
        name=name,
        csr_str=cak.csr.base64)
    assert csrr.get_resource(mock_api_client) == mock_response
    mock_read_csr.assert_called_once_with(name)
    # test the cache worked
    assert csrr._resource_cache == mock_response


@mock.patch('k8s_user.k8s.csr_resource.kubernetes.client.CertificatesV1beta1Api')
def test__csrresource__get_resource__404(mock_CertificatesV1beta1Api):

    name = "joe"
    mock_read_csr = mock.MagicMock()
    mock_read_csr.side_effect = ApiException(status = 404)
    mock_CertificatesV1beta1Api.return_value.read_certificate_signing_request_status = mock_read_csr

    mock_api_client = mock.Mock(spec=kubernetes.client.ApiClient) 
    cak = CSRandKey(common_name=name)
    csrr = CSRResource(
        name=name,
        csr_str=cak.csr.base64)
    assert csrr.get_resource(mock_api_client) is None
    mock_read_csr.assert_called_once_with(name)


@pytest.mark.parametrize("resource,expected", 
[
    pytest.param({"dummy": "resource"}, True),
    pytest.param(None, False),
])
def test__csrresource__resource_exists(resource, expected):
    name = "joe"
    mock_api_client = mock.Mock(spec=kubernetes.client.ApiClient) 
    cak = CSRandKey(common_name=name)
    csrr = CSRResource(
        name=name,
        csr_str=cak.csr.base64)
    with mock.patch.object(csrr, 'get_resource') as mock__get_resource:
        mock__get_resource.return_value = resource
        assert csrr.resource_exists(mock_api_client) == expected


@mock.patch('k8s_user.k8s.csr_resource.kubernetes.client.CertificatesV1beta1Api')
def test__csrresource__create__new(mock_CertificatesV1beta1Api):

    mock_response = {'dummy': "response"}
    mock_create_csr = mock.MagicMock()
    mock_create_csr.return_value = mock_response
    mock_CertificatesV1beta1Api.return_value.create_certificate_signing_request = mock_create_csr

    name = "joe"
    mock_api_client = mock.Mock(spec=kubernetes.client.ApiClient) 
    cak = CSRandKey(common_name=name)
    csrr = CSRResource(
        name=name,
        csr_str=cak.csr.base64)

    with mock.patch.object(csrr, 'resource_exists') as mock__resource_exists:
        mock__resource_exists.return_value = False
        assert csrr.create(mock_api_client) == mock_response
        mock_create_csr.assert_called_once()


@mock.patch('k8s_user.k8s.csr_resource.kubernetes.client.CertificatesV1beta1Api')
def test__csrresource__create__existing(mock_CertificatesV1beta1Api):

    mock_response = {'dummy': "response"}

    name = "joe"
    mock_api_client = mock.Mock(spec=kubernetes.client.ApiClient) 
    cak = CSRandKey(common_name=name)
    csrr = CSRResource(
        name=name,
        csr_str=cak.csr.base64)

    with mock.patch.object(csrr, 'resource_exists') as mock__resource_exists:
        mock__resource_exists.return_value = True
        with mock.patch.object(csrr, 'get_resource') as mock__get_resource:
            mock__get_resource.return_value = mock_response
            assert csrr.create(mock_api_client) == mock_response


@mock.patch('k8s_user.k8s.csr_resource.kubernetes.client.CertificatesV1beta1Api')
def test__csrresource__approve(mock_CertificatesV1beta1Api):
    class DummyStatus:
        conditions = None

    class DummyResponse:
        def __init__(self):
            self.status = DummyStatus()

    mock_response = DummyResponse()

    mock_approve_csr = mock.MagicMock()
    mock_approve_csr.return_value = mock_response
    mock_CertificatesV1beta1Api.return_value.replace_certificate_signing_request_approval = mock_approve_csr

    name = "joe"
    mock_api_client = mock.Mock(spec=kubernetes.client.ApiClient) 
    cak = CSRandKey(common_name=name)
    csrr = CSRResource(
        name=name,
        csr_str=cak.csr.base64)

    with mock.patch.object(csrr, 'get_resource') as mock__get_resource:
        mock__get_resource.return_value = mock_response
        assert csrr.approve(mock_api_client) == mock_response
        mock_approve_csr.assert_called_once()
        assert csrr._resource_cache == mock_response

# def test__csrresource__get_cert():

#     class DummyStatus:
#         certificate = "foo_cert"

#     class DummyResponse:
#         def __init__(self):
#             self.status = DummyStatus()

#     mock_response = DummyResponse()

#     name = "joe"
#     mock_api_client = mock.Mock(spec=kubernetes.client.ApiClient) 
#     cak = CSRandKey(common_name=name)
#     csrr = CSRResource(
#         name=name,
#         csr_str=cak.csr.base64)

#     with mock.patch.object(csrr, 'get_resource') as mock__get_resource:
#         breakpoint()
#         mock__get_resource.return_value = mock_response
#         assert csrr.approve(mock_api_client) == 'foo_cert'



from setuptools import setup, find_packages


from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    name="kubernetes-user",
    version="0.0.3",
    author="Joe Jasinski",
    description="A package for creating kubernetes users.",
    long_description=long_description,
    long_description_content_type='text/markdown',
    url="https://github.com/JoeJasinski/k8s_user",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        "kubernetes",
        "cryptography",
        "pyyaml",
    ],
    extras_require={
        "test":  ["pytest", "pytest-cov", "black", "coverage"],
    },
    entry_points={
        "console_scripts": [
            "k8s_user = k8s_user.__main__:main",
        ],
    }
)